@@ -24,6 +24,7 @@ import cgi
 import mimetypes
 import os
 import re
+import StringIO
 import sys
 import traceback
 try:
@@ -32,7 +33,7 @@ except ImportError:
     from cgi import parse_qs
 
 __author__ = 'Daniel Lindsley'
-__version__ = ('0', '6', '4')
+__version__ = ('0', '6', '7')
 __license__ = 'BSD'
 
 
@@ -100,9 +101,11 @@ class RequestError(Exception):
         super(RequestError, self).__init__(message)
         self.hide_traceback = hide_traceback
 
+
 class Forbidden(RequestError):
     status = 403
 
+
 class NotFound(RequestError):
     status = 404
     
@@ -110,9 +113,11 @@ class NotFound(RequestError):
         super(NotFound, self).__init__(message)
         self.hide_traceback = hide_traceback
 
+
 class AppError(RequestError):
     status = 500
 
+
 class Redirect(RequestError):
     """
     Redirects the user to a different URL.
@@ -129,11 +134,23 @@ class Redirect(RequestError):
         self.args = ["Redirecting to '%s'..." % self.url]
 
 
+class lazyproperty(object):
+    """A property whose value is computed only once. """
+    def __init__(self, function):
+        self._function = function
+
+    def __get__(self, obj, _=None):
+        if obj is None:
+            return self
+        
+        value = self._function(obj)
+        setattr(obj, self._function.func_name, value)
+        return value
+
+
 class Request(object):
     """An object to wrap the environ bits in a friendlier way."""
     GET = {}
-    POST = {}
-    PUT = {}
     
     def __init__(self, environ, start_response):
         self._environ = environ
@@ -152,12 +169,19 @@ class Request(object):
             pass
         
         self.GET = self.build_get_dict()
-        
-        if self.method == 'POST':
-            self.POST = self.build_complex_dict()
-        elif self.method == 'PUT':
-            self.PUT = self.build_complex_dict()
     
+    @lazyproperty
+    def POST(self):
+        return self.build_complex_dict()
+    
+    @lazyproperty
+    def PUT(self):
+        return self.build_complex_dict()
+    
+    @lazyproperty
+    def body(self):
+        """Content of the request."""
+        return self._environ['wsgi.input'].read(self.content_length)
     
     def build_get_dict(self):
         """Takes GET data and rips it apart into a dict."""
@@ -177,7 +201,7 @@ class Request(object):
     
     def build_complex_dict(self):
         """Takes POST/PUT data and rips it apart into a dict."""
-        raw_data = cgi.FieldStorage(fp=self._environ['wsgi.input'], environ=self._environ)
+        raw_data = cgi.FieldStorage(fp=StringIO.StringIO(self.body), environ=self._environ)
         query_dict = {}
         
         for field in raw_data:
@@ -211,16 +235,38 @@ class Response(object):
     
     def send(self, start_response):
         status = "%d %s" % (self.status, HTTP_MAPPINGS.get(self.status))
-        headers = [('Content-Type', self.content_type)] + self.headers
-        start_response(status, headers)
-        return self.output
+        headers = [('Content-Type', "%s; charset=utf-8" % self.content_type)] + self.headers
+        final_headers = []
+        
+        # Because Unicode is unsupported...
+        for header in headers:
+            final_headers.append((self.convert_to_ascii(header[0]), self.convert_to_ascii(header[1])))
+        
+        start_response(status, final_headers)
+        
+        if isinstance(self.output, unicode):
+            return self.output.encode('utf-8')
+        else:
+            return self.output
+    
+    def convert_to_ascii(self, data):
+        if isinstance(data, unicode):
+            try:
+                return data.encode('us-ascii')
+            except UnicodeError, e:
+                raise
+        else:
+            return str(data)
 
 
 def handle_request(environ, start_response):
     """The main handler. Dispatches to the user's code."""
     try:
         request = Request(environ, start_response)
-        
+    except Exception, e:
+        return handle_error(e)
+    
+    try:
         (re_url, url, callback), kwargs = find_matching_url(request)
         response = callback(request, **kwargs)
     except Exception, e:
@@ -232,8 +278,11 @@ def handle_request(environ, start_response):
     return response.send(start_response)
 
 
-def handle_error(exception, request):
+def handle_error(exception, request=None):
     """If an exception is thrown, deal with it and present an error page."""
+    if request is None:
+        request = {'_environ': {'PATH_INFO': ''}}
+    
     if not getattr(exception, 'hide_traceback', False):
         (e_type, e_value, e_tb) = sys.exc_info()
         message = "%s occurred on '%s': %s\nTraceback: %s" % (
@@ -489,6 +538,13 @@ def tornado_adapter(host, port):
     IOLoop.instance().start()
 
 
+def gunicorn_adapter(host, port):
+    from gunicorn.arbiter import Arbiter
+    from gunicorn.config import Config
+    arbiter = Arbiter(Config({'bind': "%s:%d" % (host, int(port)), 'workers': 4}), handle_request)
+    arbiter.run()
+
+
 WSGI_ADAPTERS = {
     'wsgiref': wsgiref_adapter,
     'appengine': appengine_adapter,
@@ -498,6 +554,7 @@ WSGI_ADAPTERS = {
     'twisted': twisted_adapter,
     'diesel': diesel_adapter,
     'tornado': tornado_adapter,
+    'gunicorn': gunicorn_adapter,
 }
 
 
